/*
** forw.c -- forward a message, or group of messages.
**
** This code is Copyright (c) 2002, by the authors of nmh.  See the
** COPYRIGHT file in the root directory of the nmh distribution for
** complete copyright information.
*/

#include <h/mh.h>
#include <fcntl.h>
#include <h/fmt_scan.h>
#include <h/tws.h>
#include <h/utils.h>


#define IFORMAT  "digest-issue-%s"
#define VFORMAT  "digest-volume-%s"

static struct swit switches[] = {
#define ANNOSW  0
	{ "annotate", 0 },
#define NANNOSW  1
	{ "noannotate", 0 },
#define EDITRSW  2
	{ "editor editor", 0 },
#define FORMSW  3
	{ "form formfile", 0 },
#define DGSTSW  4
	{ "digest list", 0 },
#define ISSUESW  5
	{ "issue number", 0 },
#define VOLUMSW  6
	{ "volume number", 0 },
#define WHATSW  7
	{ "whatnowproc program", 0 },
#define VERSIONSW  8
	{ "Version", 0 },
#define HELPSW  9
	{ "help", 0 },
#define BILDSW  12
	{ "build", 5 },  /* interface from mhe */
	{ NULL, 0 }
};

static char drft[BUFSIZ];
static struct msgs *mp = NULL;


/*
** static prototypes
*/
static void add_forw_hdr(char *);
static int build_form(char *, char *, int, int);


int
main(int argc, char **argv)
{
	int msgp = 0, anot = 0;
	int issue = 0, volume = 0;
	int in;
	int out, msgnum;
	char *cp, *cwd, *maildir;
	char *digest = NULL, *ed = NULL;
	char *folder = NULL;
	char *form = NULL, buf[BUFSIZ], value[10];
	char **argp, **arguments, *msgs[MAXARGS];
	char *fmtstr;
	int buildsw = 0;

#ifdef LOCALE
	setlocale(LC_ALL, "");
#endif
	invo_name = mhbasename(argv[0]);

	/* read user profile/context */
	context_read();

	arguments = getarguments(invo_name, argc, argv, 1);
	argp = arguments;

	while ((cp = *argp++)) {
		if (*cp == '-') {
			switch (smatch(++cp, switches)) {
			case AMBIGSW:
				ambigsw(cp, switches);
				done(1);
			case UNKWNSW:
				adios(NULL, "-%s unknown", cp);

			case HELPSW:
				snprintf(buf, sizeof(buf), "%s [+folder] [msgs] [switches]", invo_name);
				print_help(buf, switches, 1);
				done(1);
			case VERSIONSW:
				print_version(invo_name);
				done(1);

			case ANNOSW:
				anot++;
				continue;
			case NANNOSW:
				anot = 0;
				continue;

			case EDITRSW:
				if (!(ed = *argp++) || *ed == '-')
					adios(NULL, "missing argument to %s",
							argp[-2]);
				continue;

			case WHATSW:
				if (!(whatnowproc = *argp++) ||
						*whatnowproc == '-')
					adios(NULL, "missing argument to %s",
							argp[-2]);
				continue;

			case BILDSW:
				buildsw++;
				continue;

			case FORMSW:
				if (!(form = *argp++) || *form == '-')
					adios(NULL, "missing argument to %s",
							argp[-2]);
				continue;

			case DGSTSW:
				if (!(digest = *argp++) || *digest == '-')
					adios(NULL, "missing argument to %s",
							argp[-2]);
				//mime = 0;
				continue;
			case ISSUESW:
				if (!(cp = *argp++) || *cp == '-')
					adios(NULL, "missing argument to %s",
							argp[-2]);
				if ((issue = atoi(cp)) < 1)
					adios(NULL, "bad argument %s %s",
							argp[-2], cp);
				continue;
			case VOLUMSW:
				if (!(cp = *argp++) || *cp == '-')
					adios(NULL, "missing argument to %s",
							argp[-2]);
				if ((volume = atoi(cp)) < 1)
					adios(NULL, "bad argument %s %s",
							argp[-2], cp);
				continue;
			}
		}
		if (*cp == '+' || *cp == '@') {
			if (folder)
				adios(NULL, "only one folder at a time!");
			else
				folder = getcpy(expandfol(cp));
		} else {
			msgs[msgp++] = cp;
		}
	}

	cwd = getcpy(pwd());
	strncpy(drft, buildsw ? toabsdir("draft") : m_draft(seq_beyond),
			sizeof(drft));
	/*
	** FIXME: (concerning MHE support (buildsw) only)
	** There's no check if the draft already exists. mmh has removed
	** this case by having the draft folder. I won't add code only to
	** handle this legacy issue for MHE. -- meillo@marmaro.de 2012-05
	*/

	/*
	** Forwarding a message.
	*/
	if (!msgp)
		msgs[msgp++] = seq_cur;
	if (!folder)
		folder = getcurfol();
	maildir = toabsdir(folder);

	if (chdir(maildir) == NOTOK)
		adios(maildir, "unable to change directory to");

	/* read folder and create message structure */
	if (!(mp = folder_read(folder)))
		adios(NULL, "unable to read folder %s", folder);

	/* check for empty folder */
	if (mp->nummsg == 0)
		adios(NULL, "no messages in %s", folder);

	/* parse all the message ranges/sequences and set SELECTED */
	for (msgnum = 0; msgnum < msgp; msgnum++)
		if (!m_convert(mp, msgs[msgnum]))
			done(1);
	seq_setprev(mp);  /* set the previous sequence */

	if ((out = creat(drft, m_gmprot())) == NOTOK)
		adios(drft, "unable to create");

	/* Open form (component) file. */
	if (digest) {
		if (issue == 0) {
			snprintf(buf, sizeof(buf), IFORMAT, digest);
			if (volume == 0 && (cp = context_find(buf))
					&& ((issue = atoi(cp)) < 0))
				issue = 0;
			issue++;
		}
		if (volume == 0) {
			snprintf(buf, sizeof(buf), VFORMAT, digest);
			if ((cp = context_find(buf)) == NULL ||
					(volume = atoi(cp)) <= 0)
				volume = 1;
		}
		if (!form)
			form = digestcomps;
		in = build_form(form, digest, volume, issue);
		cpydata(in, out, form, drft);
		close(in);
	} else {
		fmtstr = new_fs(form, forwcomps);
		if (write(out, fmtstr, strlen(fmtstr)) != (int)strlen(fmtstr)) {
			adios(drft, "error writing");
		}
	}
	close(out);

	add_forw_hdr(drft);

	if (digest) {
		snprintf(buf, sizeof(buf), IFORMAT, digest);
		snprintf(value, sizeof(value), "%d", issue);
		context_replace(buf, getcpy(value));
		snprintf(buf, sizeof(buf), VFORMAT, digest);
		snprintf(value, sizeof(value), "%d", volume);
		context_replace(buf, getcpy(value));
	}

	context_replace(curfolder, folder); /* update current folder */
	seq_setcur(mp, mp->lowsel);  /* update current message */
	seq_save(mp);  /* synchronize sequences */
	context_save();  /* save the context file */

	if (buildsw)
		done(0);
	what_now(ed, NOUSE, drft, NULL, 0, mp,
		anot ? "Forwarded" : NULL, cwd);
	done(1);
	return 1;
}


/*
** Create an attachment header for the to be forward messages.
*/
static void
add_forw_hdr(char *draft)
{
	int msgnum;
	char buffer[BUFSIZ];

	snprintf(buffer, sizeof buffer, "anno -append -nodate '%s' "
			"-comp '%s' -text '+%s",
			draft, attach_hdr, mp->foldpath);
	for (msgnum = mp->lowsel; msgnum <= mp->hghsel; msgnum++) {
		if (!is_selected(mp, msgnum)) {
			continue;
		}
		/* TODO: Check for buffer length! */
		strcat(buffer, " ");
		strcat(buffer, m_name(msgnum));
	}
	strcat(buffer, "'");
	/* TODO: This check is bad, but better than nothing */
	if (strlen(buffer) > BUFSIZ) {
		adios(NULL, "Too long attachment header line. Forward less messages.");
	}
	if (system(buffer) != 0) {
		advise(NULL, "unable to add attachment header");
	}
}


static int
build_form(char *form, char *digest, int volume, int issue)
{
	int in;
	int fmtsize;
	register char *fmtstr;
	char *line, tmpfil[BUFSIZ];
	FILE *tmp;
	register struct comp *cptr;
	struct format *fmt;
	int dat[5];
	char *cp = NULL;

	/* Get new format string */
	fmtstr = new_fs(form, NULL);
	fmtsize = strlen(fmtstr) + 256;

	/* Compile format string */
	fmt_compile(fmtstr, &fmt);

	FINDCOMP(cptr, "digest");
	if (cptr)
		cptr->c_text = digest;
	FINDCOMP(cptr, "date");
	if (cptr)
		cptr->c_text = getcpy(dtimenow());

	dat[0] = issue;
	dat[1] = volume;
	dat[2] = 0;
	dat[3] = fmtsize;
	dat[4] = 0;

	cp = m_mktemp2(NULL, invo_name, NULL, &tmp);
	if (cp == NULL) adios("forw", "unable to create temporary file");
	strncpy(tmpfil, cp, sizeof(tmpfil));
	unlink(tmpfil);
	if ((in = dup(fileno(tmp))) == NOTOK)
		adios("dup", "unable to");

	line = mh_xmalloc((unsigned) fmtsize);
	fmt_scan(fmt, line, fmtsize, dat);
	fputs(line, tmp);
	free(line);
	if (fclose(tmp))
		adios(tmpfil, "error writing");

	lseek(in, (off_t) 0, SEEK_SET);
	return in;
}
