/*
** whatnow.c -- the WhatNow shell
**
** This code is Copyright (c) 2002, by the authors of nmh.  See the
** COPYRIGHT file in the root directory of the nmh distribution for
** complete copyright information.
**
**  The inclusion of attachments is eased by
**  using the header field name mechanism added to anno and send.
**  The header field name for attachments is predefined.
**
**  Several commands have been added at the whatnow prompt:
**
**        cd [ directory ]        This option works just like the shell's
**                                cd command and lets the user change the
**                                directory from which attachments are
**                                taken so that long path names are not
**                                needed with every file.
**
**        ls [ ls-options ]       This option works just like the normal
**                                ls command and exists to allow the user
**                                to verify file names in the directory.
**
**        pwd                     This option works just like the normal
**                                pwd command and exists to allow the user
**                                to verify the directory.
**
**        attach files            This option attaches the named files to
**                                the draft.
**
**        alist                   This option lists the attachments on the
**                                draft.
**
**        detach files            This option removes attachments from the
**        detach -n numbers       draft.  This can be done by file name or
**                                by attachment number.
*/

#include <h/mh.h>
#include <fcntl.h>
#include <signal.h>
#include <h/mime.h>
#include <h/utils.h>

static struct swit whatnowswitches[] = {
#define EDITRSW  0
	{ "editor editor", 0 },
#define PRMPTSW  1
	{ "prompt string", 4 },
#define VERSIONSW  2
	{ "Version", 0 },
#define HELPSW  3
	{ "help", 0 },
	{ NULL, 0 }
};

/*
** Options at the "whatnow" prompt
*/
static struct swit aleqs[] = {
#define EDITSW  0
	{ "edit [<editor> <switches>]", 0 },
#define REFILEOPT  1
	{ "refile [<switches>] +folder", 0 },
#define DISPSW  2
	{ "display", 0 },
#define LISTSW  3
	{ "list", 0 },
#define SENDSW  4
	{ "send [<switches>]", 0 },
#define QUITSW  5
	{ "quit", 0 },
#define DELETESW  6
	{ "delete", 0 },
#define CDCMDSW  7
	{ "cd [directory]", 0 },
#define PWDCMDSW  8
	{ "pwd", 0 },
#define LSCMDSW  9
	{ "ls", 0 },
#define ATTACHCMDSW  10
	{ "attach", 0 },
#define DETACHCMDSW  11
	{ "detach [-n]", 0 },
#define ALISTCMDSW  12
	{ "alist", 0 },
	{ NULL, 0 }
};

static char *myprompt = "\nWhat now? ";

/*
** static prototypes
*/
static int editfile(char **, char **, char *, int, struct msgs *,
	char *, char *);
static int sendfile(char **, char *);
static int refile(char **, char *);
static int removefile(char *);
static void writelscmd(char *, int, char **);
static void writesomecmd(char *buf, int bufsz, char *cmd, char *trailcmd, char **argp);
static FILE* popen_in_dir(const char *dir, const char *cmd, const char *type);
static int system_in_dir(const char *dir, const char *cmd);


#ifdef HAVE_LSTAT
static int copyf(char *, char *);
#endif


int
main(int argc, char **argv)
{
	int use = 0;
	char *cp;
	char *ed = NULL, *drft = NULL, *msgnam = NULL;
	char buf[BUFSIZ], prompt[BUFSIZ];
	char **argp, **arguments;
	struct stat st;
	char cwd[MAXPATHLEN + 1];  /* current working directory */
	char file[MAXPATHLEN + 1];  /* file name buffer */
	char shell[MAXPATHLEN + 1];  /* shell response buffer */
	FILE *f;  /* read pointer for bgnd proc */
	int n;  /* set on -n to detach command */

#ifdef LOCALE
	setlocale(LC_ALL, "");
#endif

	invo_name = mhbasename(argv[0]);

	/* read user profile/context */
	context_read();

	arguments = getarguments(invo_name, argc, argv, 1);
	argp = arguments;

	/*
	** Get the initial current working directory.
	*/

	if (!getcwd(cwd, sizeof (cwd))) {
		adios("getcwd", "could not get working directory");
	}

	while ((cp = *argp++)) {
		if (*cp == '-') {
			switch (smatch(++cp, whatnowswitches)) {
			case AMBIGSW:
				ambigsw(cp, whatnowswitches);
				done(1);
			case UNKWNSW:
				adios(NULL, "-%s unknown", cp);

			case HELPSW:
				snprintf(buf, sizeof(buf),
						"%s [switches] [file]",
						invo_name);
				print_help(buf, whatnowswitches, 1);
				done(1);
			case VERSIONSW:
				print_version(invo_name);
				done(1);

			case EDITRSW:
				if (!(ed = *argp++) || *ed == '-')
					adios(NULL, "missing argument to %s",
							argp[-2]);
				continue;

			case PRMPTSW:
				if (!(myprompt = *argp++) || *myprompt == '-')
					adios(NULL, "missing argument to %s",
							argp[-2]);
				continue;

			}
		}
		if (drft)
			adios(NULL, "only one draft at a time!");
		else
			drft = cp;
	}

	if ((!drft && !(drft = getenv("mhdraft"))) || !*drft)
		drft = getcpy(m_draft(seq_cur));

	msgnam = (cp = getenv("mhaltmsg")) && *cp ? getcpy(cp) : NULL;

	if ((cp = getenv("mhuse")) && *cp)
		use = atoi(cp);

	if (!ed && !(ed = getenv("mheditor"))) {
		ed = "";  /* Don't initially edit the draft */
	}

	/* start editing the draft, unless editor is the empty string */
	if (*ed) {
		if (editfile(&ed, NULL, drft, use, NULL, msgnam, NULL) <0) {
			if (!use) {
				unlink(drft);
			}
			advise(NULL, "Try again.");
			done(1);
		}
	}

	snprintf(prompt, sizeof(prompt), myprompt, invo_name);
	for (;;) {
		if (!(argp = getans(prompt, aleqs))) {
			unlink(altmsglink);
			done(1);
		}
		switch (smatch(*argp, aleqs)) {
		case DISPSW:
			/* display the msg being replied to or distributed */
			if (msgnam) {
				snprintf(buf, sizeof buf, "%s '%s'",
						listproc, msgnam);
				system(buf);
			} else {
				advise(NULL, "no alternate message to display");
			}
			break;

		case EDITSW:
			/* Call an editor on the draft file */
			if (*++argp)
				ed = *argp++;
			editfile(&ed, argp, drft, NOUSE, NULL, msgnam, NULL);
			break;

		case LISTSW:
			/* display the draft file */
			snprintf(buf, sizeof buf, "%s '%s'", listproc, drft);
			system(buf);
			break;

		case QUITSW:
			/* quit */
			if (stat(drft, &st) != NOTOK) {
				advise(NULL, "draft left on %s", drft);
			}
			done(1);

		case DELETESW:
			/* Delete draft and exit */
			removefile(drft);
			done(1);

		case SENDSW:
			/* Send draft */
			sendfile(++argp, drft);
			break;

		case REFILEOPT:
			/* Refile the draft */
			if (refile(++argp, drft) == 0) {
				done(0);
			}
			break;

		case CDCMDSW:
			/*
			** Change the working directory for attachments
			**
			** Run the directory through the user's shell
			** so that we can take advantage of any syntax
			** that the user is accustomed to.  Read back
			** the absolute path.
			*/

			if (*(argp+1) == NULL) {
				sprintf(buf, "$SHELL -c \"cd;pwd\"");
			} else {
				writesomecmd(buf, BUFSIZ, "cd", "pwd", argp);
			}
			if ((f = popen_in_dir(cwd, buf, "r"))) {
				fgets(cwd, sizeof (cwd), f);

				if (strchr(cwd, '\n'))
					*strchr(cwd, '\n') = '\0';

				pclose(f);
			} else {
				advise("popen", "could not get directory");
			}

			break;

		case PWDCMDSW:
			/* Print the working directory for attachments */
			printf("%s\n", cwd);
			break;

		case LSCMDSW:
			/*
			** List files in the current attachment working
			** directory
			**
			** Use the user's shell so that we can take
			** advantage of any syntax that the user is
			** accustomed to.
			*/
			writelscmd(buf, sizeof(buf), argp);
			system_in_dir(cwd, buf);
			break;

		case ALISTCMDSW:
			/*
			** List attachments on current draft.
			*/
			snprintf(buf, sizeof buf, "anno -list -comp '%s' "
					"-number all -text IGNORE '%s'",
					attach_hdr, drft);
			if (system(buf) != 0) {
				advise(NULL, "Could not list attachment headers.");
			}
			break;

		case ATTACHCMDSW:
			/*
			** Attach files to current draft.
			*/

			if (*(argp+1) == NULL) {
				advise(NULL, "attach command requires file argument(s).");
				break;
			}

			/*
			** Build a command line that causes the user's
			** shell to list the file name arguments.
			** This handles and wildcard expansion, tilde
			** expansion, etc.
			*/
			writelscmd(buf, sizeof(buf), argp);

			/*
			** Read back the response from the shell,
			** which contains a number of lines with one
			** file name per line.  Remove off the newline.
			** Determine whether we have an absolute or
			** relative path name.  Prepend the current
			** working directory to relative path names.
			** Add the attachment annotation to the draft.
			*/
			if ((f = popen_in_dir(cwd, buf, "r"))) {
				char buf[BUFSIZ];

				while (fgets(shell, sizeof(shell), f)) {
					*(strchr(shell, '\n')) = '\0';

					if (*shell == '/')
						sprintf(file, "%s", shell);
					else {
						sprintf(file, "%s/%s", cwd,
								shell);
					}
					snprintf(buf, sizeof buf,
							"anno -nodate -append "
							"-comp '%s' -text '%s'"
							" '%s'",
							attach_hdr, file,
							drft);
					if (system(buf) != 0) {
						advise(NULL, "Could not add attachment header.");
					}
				}

				pclose(f);
			} else {
				advise("popen", "could not get file from shell");
			}

			break;

		case DETACHCMDSW:
			/*
			** Detach files from current draft.
			**
			** Scan the arguments for a -n.  Mixed file
			** names and numbers aren't allowed, so this
			** catches a -n anywhere in the argument list.
			*/
			for (n = 0, arguments = argp + 1;
					*arguments != NULL;
					arguments++) {
				if (strcmp(*arguments, "-n") == 0) {
						n = 1;
						break;
				}
			}

			/*
			** A -n was found so interpret the arguments as
			** attachment numbers.  Decrement any remaining
			** argument number that is greater than the one
			** just processed after processing each one so
			** that the numbering stays correct.
			*/
			if (n == 1) {
				for (arguments=argp+1; *arguments;
						arguments++) {
					if (strcmp(*arguments, "-n") == 0)
						continue;

					if (**arguments != '\0') {
						char buf[BUFSIZ];

						n = atoi(*arguments);
						snprintf(buf, sizeof buf, "anno -delete -comp '%s' -number '%d' '%s'", attach_hdr, n, drft);
						if (system(buf) != 0) {
							advise(NULL, "Could not delete attachment header.");
						}

						for (argp=arguments+1; *argp;
								argp++) {
							if (atoi(*argp) > n) {
								if (atoi(*argp) == 1)
									*argp = "";
								else
									sprintf(*argp, "%d", atoi(*argp) - 1);
							}
						}
					}
				}
				break;
			}
			/* else */

			/*
			** The arguments are interpreted as file names.
			** Run them through the user's shell for wildcard
			** expansion and other goodies.  Do this from
			** the current working directory if the argument
			** is not an absolute path name (does not begin
			** with a /).
			**
			** We feed all the file names to the shell at
			** once, otherwise you can't provide a file name
			** with a space in it.
			*/
			writelscmd(buf, sizeof(buf), argp);
			if ((f = popen_in_dir(cwd, buf, "r"))) {
				while (fgets(shell, sizeof (shell), f)) {
					*(strchr(shell, '\n')) = '\0';
					snprintf(buf, sizeof buf,
							"anno -delete -comp "
							"'%s' -text '%s' '%s'",
							attach_hdr, shell,
							drft);
					if (system(buf) != 0) {
						advise(NULL, "Could not delete attachment header.");
					}
				}
				pclose(f);
			} else {
				advise("popen", "could not get file from shell");
			}

			break;

		default:
			/* Unknown command */
			advise(NULL, "say what?");
			break;
		}
	}
	/*NOTREACHED*/
}



/*
** Build a command line of the form $SHELL -c "cd 'cwd'; cmd argp ... ;
** trailcmd".
*/
static void
writesomecmd(char *buf, int bufsz, char *cmd, char *trailcmd, char **argp)
{
	char *cp;
	/*
	** Note that we do not quote -- the argp from the user
	** is assumed to be quoted as they desire. (We can't treat
	** it as pure literal as that would prevent them using ~,
	** wildcards, etc.) The buffer produced by this function
	** should be given to popen_in_dir() or system_in_dir() so
	** that the current working directory is set correctly.
	*/
	int ln = snprintf(buf, bufsz, "$SHELL -c \"%s", cmd);
	/*
	** NB that some snprintf() return -1 on overflow rather than the
	** new C99 mandated 'number of chars that would have been written'
	*/
	/*
	** length checks here and inside the loop allow for the
	** trailing ';', trailcmd, '"' and NUL
	*/
	int trailln = strlen(trailcmd) + 3;
	if (ln < 0 || ln + trailln > bufsz)
		adios(NULL, "arguments too long");

	cp = buf + ln;

	while (*++argp) {
		ln = strlen(*argp);
		/* +1 for leading space */
		if (ln + trailln + 1 > bufsz - (cp-buf))
			adios(NULL, "arguments too long");
		*cp++ = ' ';
		memcpy(cp, *argp, ln+1);
		cp += ln;
	}
	if (*trailcmd) {
		*cp++ = ';';
		strcpy(cp, trailcmd);
		cp += trailln - 3;
	}
	*cp++ = '"';
	*cp = 0;
}

/*
** Build a command line that causes the user's shell to list the file name
** arguments.  This handles and wildcard expansion, tilde expansion, etc.
*/
static void
writelscmd(char *buf, int bufsz, char **argp)
{
	writesomecmd(buf, bufsz, "ls", "", argp);
}

/*
** Like system(), but run the command in directory dir.
** This assumes the program is single-threaded!
*/
static int
system_in_dir(const char *dir, const char *cmd)
{
	char olddir[BUFSIZ];
	int r;
	if (getcwd(olddir, sizeof(olddir)) == 0)
		adios("getcwd", "could not get working directory");
	if (chdir(dir) != 0)
		adios("chdir", "could not change working directory");
	r = system(cmd);
	if (chdir(olddir) != 0)
		adios("chdir", "could not change working directory");
	return r;
}

/* ditto for popen() */
static FILE*
popen_in_dir(const char *dir, const char *cmd, const char *type)
{
	char olddir[BUFSIZ];
	FILE *f;
	if (getcwd(olddir, sizeof(olddir)) == 0)
		adios("getcwd", "could not get working directory");
	if (chdir(dir) != 0)
		adios("chdir", "could not change working directory");
	f = popen(cmd, type);
	if (chdir(olddir) != 0)
		adios("chdir", "could not change working directory");
	return f;
}


/*
** EDIT
*/

static char *edsave = NULL;  /* the editor we used previously */


static int
editfile(char **ed, char **arg, char *file, int use, struct msgs *mp,
	char *altmsg, char *cwd)
{
	int pid, status, vecp;
	char altpath[BUFSIZ], linkpath[BUFSIZ];
	char *cp, *vec[MAXARGS];
	struct stat st;

#ifdef HAVE_LSTAT
	int slinked = 0;
#endif /* HAVE_LSTAT */

	if (!*ed || !**ed) {
		/* We have no explicit editor. */
		if (edsave) {
			/* Use the previous editor ... */
			*ed = edsave;
			if (!(cp = mhbasename(*ed)))
				cp = *ed;

			/* but prefer one specified via "editor-next" */
			cp = concat(cp, "-next", NULL);
			if ((cp = context_find(cp)))
				*ed = cp;
		} else {
			/* set initial editor */
			*ed = defaulteditor;
		}
	}

	if (altmsg) {
		if (!mp || *altmsg == '/' || !cwd)
			strncpy(altpath, altmsg, sizeof(altpath));
		else
			snprintf(altpath, sizeof(altpath), "%s/%s",
					mp->foldpath, altmsg);
		if (!cwd)
			strncpy(linkpath, altmsglink, sizeof(linkpath));
		else
			snprintf(linkpath, sizeof(linkpath), "%s/%s",
					cwd, altmsglink);
		unlink(linkpath);
#ifdef HAVE_LSTAT
		if (link(altpath, linkpath) == NOTOK) {
			symlink(altpath, linkpath);
			slinked = 1;
		} else {
			slinked = 0;
		}
#else /* not HAVE_LSTAT */
		link(altpath, linkpath);
#endif /* not HAVE_LSTAT */
	}

	context_save();  /* save the context file */
	fflush(stdout);

	switch (pid = fork()) {
	case NOTOK:
		advise("fork", "unable to");
		status = NOTOK;
		break;

	case OK:
		if (cwd)
			chdir(cwd);
		if (altmsg) {
			if (mp)
				m_putenv("mhfolder", mp->foldpath);
			m_putenv("editalt", altpath);
		}

		vecp = 0;
		vec[vecp++] = mhbasename(*ed);
		if (arg)
			while (*arg)
				vec[vecp++] = *arg++;
		vec[vecp++] = file;
		vec[vecp] = NULL;

		execvp(*ed, vec);
		fprintf(stderr, "%s: unable to exec ", invo_name);
		perror(*ed);
		_exit(-1);

	default:
		if ((status = pidwait(pid, NOTOK))) {
			if ((status & 0xff00) == 0xff00) {
				/* cmd not found or pidwait() failed */
				status = -1;
				break;
			}
			if (status & 0x00ff) {
				/* terminated by signal */
				advise(NULL, "%s terminated by signal %d",
						*ed, status & 0x7f);
			} else {
				/* failure exit */
				advise(NULL, "%s exited with return code %d",
						*ed, (status & 0xff00) >> 8);
			}
			status = -1;
			break;
		}

#ifdef HAVE_LSTAT
		if (altmsg && mp && !is_readonly(mp) && (slinked ?
				lstat (linkpath, &st) != NOTOK &&
				S_ISREG(st.st_mode) &&
				copyf(linkpath, altpath) == NOTOK :
				stat(linkpath, &st) != NOTOK &&
				st.st_nlink == 1 &&
				(unlink(altpath) == NOTOK ||
				link(linkpath, altpath) == NOTOK)))
			advise(linkpath, "unable to update %s from", altmsg);
#else /* HAVE_LSTAT */
		if (altmsg && mp && !is_readonly(mp) &&
				stat(linkpath, &st) != NOTOK &&
				st.st_nlink == 1 &&
				(unlink(altpath) == NOTOK ||
				link(linkpath, altpath) == NOTOK))
			advise(linkpath, "unable to update %s from", altmsg);
#endif /* HAVE_LSTAT */
	}

	/* remember which editor we used */
	edsave = getcpy(*ed);

	*ed = NULL;
	if (altmsg) {
		unlink(linkpath);
	}

	return status;
}


#ifdef HAVE_LSTAT
static int
copyf(char *ifile, char *ofile)
{
	int i, in, out;
	char buffer[BUFSIZ];

	if ((in = open(ifile, O_RDONLY)) == NOTOK)
		return NOTOK;
	if ((out = open(ofile, O_WRONLY | O_TRUNC)) == NOTOK) {
		admonish(ofile, "unable to open and truncate");
		close(in);
		return NOTOK;
	}

	while ((i = read(in, buffer, sizeof(buffer))) > OK)
		if (write(out, buffer, i) != i) {
			advise(ofile, "may have damaged");
			i = NOTOK;
			break;
		}

	close(in);
	close(out);
	return i;
}
#endif /* HAVE_LSTAT */


/*
** SEND
*/

static int
sendfile(char **arg, char *file)
{
	pid_t child_id;
	int vecp;
	char *vec[MAXARGS];

	context_save();  /* save the context file */
	fflush(stdout);

	switch (child_id = fork()) {
	case NOTOK:
		advise(NULL, "unable to fork, so sending directly...");
		/* fall */
	case OK:
		vecp = 0;
		vec[vecp++] = invo_name;
		if (arg)
			while (*arg)
				vec[vecp++] = *arg++;
		vec[vecp++] = file;
		vec[vecp] = NULL;

		execvp("send", vec);
		fprintf(stderr, "%s: unable to exec ", invo_name);
		perror("send");
		_exit(-1);

	default:
		if (pidwait(child_id, OK) == 0)
			done(0);
		return 1;
	}
}


/*
** refile msg into another folder
*/
static int
refile(char **arg, char *file)
{
	pid_t pid;
	register int vecp;
	char *vec[MAXARGS];

	vecp = 0;
	vec[vecp++] = "refile";
	vec[vecp++] = "-nolink";  /* override bad .mh_profile defaults */
	vec[vecp++] = "-file";
	vec[vecp++] = file;

	while (arg && *arg) {
		vec[vecp++] = *arg++;
	}
	vec[vecp] = NULL;

	context_save();  /* save the context file */
	fflush(stdout);

	switch (pid = fork()) {
	case -1:
		advise("fork", "unable to");
		return -1;

	case 0:
		execvp(*vec, vec);
		fprintf(stderr, "%s: unable to exec ", invo_name);
		perror(*vec);
		_exit(-1);

	default:
		return (pidwait(pid, -1));
	}
}


/*
** Remove the draft file
*/

static int
removefile(char *drft)
{
	if (unlink(drft) == NOTOK)
		adios(drft, "unable to unlink");

	return OK;
}
